
class TonSdkException(Exception):
    """
    Base class for tonsdk exceptions.
    Subclasses should provide `.default_detail` properties.
    """
    default_detail = 'TonSdk error.'

    def __init__(self, detail=None):
        self.detail = self.default_detail if detail is None else detail

    def __str__(self):
        return str(self.detail)



import decimal
import json
from decimal import localcontext
from typing import Optional, Union

import aiohttp
import pkg_resources
from tronpy import Tron
from tronpy.keys import PrivateKey

from aiotx.clients._base_client import BlockMonitor
from aiotx.clients._evm_base_client import AioTxEVMBaseClient
from aiotx.exceptions import InvalidArgumentError, RpcConnectionError
from aiotx.types import BlockParam

units = {
    "sun": 1,
    "trx": 10 ** 6,
}

MIN_SUN = 1
MAX_SUN = 10 ** 18


class AioTxTRONClient(AioTxEVMBaseClient):
    def __init__(
        self, node_url, 
    ):
        super().__init__(node_url)
        self.monitor = TronMonitor(self)
        self._monitoring_task = None
        trc20_abi_json = pkg_resources.resource_string('aiotx.utils', 'trc20_abi.json')
        self._trc20_abi = json.loads(trc20_abi_json)

    def _get_abi_entries(self):
        return [entry for entry in self._trc20_abi]
    
    def generate_address(self):
        client = Tron()
        return client.generate_address()
    
    def get_address_from_private_key(self, private_key: str):
        client = Tron()
        priv_key = PrivateKey(bytes.fromhex(private_key))
        return client.generate_address(priv_key)
    
    def hex_address_to_base58(self, hex_address: str) -> str:
        # HACK sometimes we have address with 0x prefix? 
        # Should we handle it somehow?
        if hex_address.startswith("0x"):
            hex_address = hex_address.replace("0x", "41")
        client = Tron()
        if not client.is_hex_address(hex_address):
            raise TypeError("Please provide hex address")
        return client.to_base58check_address(hex_address)
    
    def base58_to_hex_address(self, address) -> str:
        client = Tron()
        if not client.is_base58check_address(address):
            raise TypeError("Please provide base58 address")
        return client.to_hex_address(address)
    
    def to_sun(self, number: Union[int, float, str, decimal.Decimal], unit: str = "trx") -> int:
        """
        Takes a number of a unit and converts it to Sun (the smallest unit of TRX).
        """
        unit = unit.lower()
        if unit not in units:
            raise ValueError(f"Unknown unit. Must be one of {'/'.join(units.keys())}")

        if isinstance(number, int) or isinstance(number, str):
            d_number = decimal.Decimal(value=number)
        elif isinstance(number, float):
            d_number = decimal.Decimal(value=str(number))
        elif isinstance(number, decimal.Decimal):
            d_number = number
        else:
            raise TypeError("Unsupported type. Must be one of integer, float, or string")

        s_number = str(number)
        unit_value = units[unit]

        if d_number == decimal.Decimal(0):
            return 0

        if d_number < 1 and "." in s_number:
            with localcontext() as ctx:
                multiplier = len(s_number) - s_number.index(".") - 1
                ctx.prec = multiplier
                d_number = decimal.Decimal(value=number, context=ctx) * 10 ** multiplier
                unit_value /= 10 ** multiplier

        with localcontext() as ctx:
            ctx.prec = 999
            result_value = decimal.Decimal(value=d_number, context=ctx) * unit_value

        if result_value < MIN_SUN or result_value > MAX_SUN:
            raise ValueError("Resulting Sun value must be between 1 and 10^18")

        return int(result_value)

    def from_sun(self, number: Union[int, str], unit: str = "trx") -> decimal.Decimal:
        """
        Converts a value in Sun (the smallest unit of TRX) to the specified unit.
        """
        if isinstance(number, str):
            if number.startswith("0x"):
                number = int(number, 16)
            else:
                number = int(number)

        unit = unit.lower()
        if unit not in units:
            raise ValueError(f"Unknown unit. Must be one of {'/'.join(units.keys())}")

        unit_value = units[unit]
        result_value = decimal.Decimal(number) / unit_value

        return result_value

    async def get_balance(self, address, block_parameter: BlockParam = BlockParam.LATEST) -> int:
        client = Tron()
        if client.is_base58check_address(address):
            address = self.base58_to_hex_address(address)
        return await super().get_balance(address, block_parameter)

    async def get_contract_balance(
        self, address, contract_address, block_parameter: BlockParam = BlockParam.LATEST
    ) -> int:
        client = Tron()
        if client.is_base58check_address(address):
            address = self.base58_to_hex_address(address)
        if client.is_base58check_address(contract_address):
            contract_address = self.base58_to_hex_address(contract_address)
        return await super().get_contract_balance(address, contract_address, block_parameter)
    
    async def get_contract_decimals(self, address: str):
        client = Tron()
        if client.is_base58check_address(address):
            address = self.base58_to_hex_address(address)
        return await super().get_contract_decimals(address)
    
    async def _make_rpc_call(self, payload, path="/jsonrpc") -> dict:
        payload["jsonrpc"] = "2.0"
        payload["id"] = 1
        payload_json = json.dumps(payload)
        headers = {"Content-Type": "application/json"}
        async with aiohttp.ClientSession() as session:
            async with session.post(self.node_url + path, data=payload_json, headers=headers) as response:
                response_text = await response.text()
                if response.status != 200:
                    raise RpcConnectionError(f"Node response status code: {response.status} response test: {response_text}")
                result = await response.json()
                if "error" not in result.keys():
                    return result["result"]
                error_code = result["error"]["code"]
                error_message = result["error"]["message"]
                if (
                        "invalid characters encountered in Hex string" in error_message or
                        "invalid hash value" in error_message or
                        "invalid address hash value" in error_message
                    ):
                        raise InvalidArgumentError(error_message)
                else:
                    raise RpcConnectionError(f"Error {error_code}: {error_message}")


class TronMonitor(BlockMonitor):
    def __init__(self, client: AioTxTRONClient, last_block: Optional[int] = None):
        self.client = client
        self.block_handlers = []
        self.transaction_handlers = []
        self.running = False
        self._last_block = last_block

    async def poll_blocks(
        self,
    ):
        network_last_block = await self.client.get_last_block_number()
        target_block = network_last_block if self._latest_block is None else self._latest_block
        if target_block > network_last_block:
            return
        block_data = await self.client.get_block_by_number(target_block)
        await self.process_transactions(block_data["transactions"])
        await self.process_block(target_block)
        self._latest_block = target_block + 1

    async def process_block(self, block):
        for handler in self.block_handlers:
            await handler(block)

    async def process_transactions(self, transactions):
        for transaction in transactions:
            transaction["aiotx_decoded_input"] = self.client.decode_transaction_input(transaction["input"])
            for handler in self.transaction_handlers:
                await handler(transaction)




from ._keystore import generate_keystore_key, generate_new_keystore
from ._mnemonic import mnemonic_is_valid, mnemonic_new, mnemonic_to_wallet_key
from ._utils import private_key_to_public_key, verify_sign

__all__ = [
    'mnemonic_new',
    'mnemonic_to_wallet_key',
    'mnemonic_is_valid',

    'generate_new_keystore',
    'generate_keystore_key',
    'private_key_to_public_key',
    'verify_sign',
    'generate_key_pair',
]





from ....boc import Cell
from ....utils import Address
from ... import Contract
from ..nft.nft_utils import create_offchain_uri_cell


class JettonMinter(Contract):
    code = 'B5EE9C7241020B010001ED000114FF00F4A413F4BCF2C80B0102016202030202CC040502037A60090A03EFD9910E38048ADF068698180B8D848ADF07D201800E98FE99FF6A2687D007D206A6A18400AA9385D47181A9AA8AAE382F9702480FD207D006A18106840306B90FD001812881A28217804502A906428027D012C678B666664F6AA7041083DEECBEF29385D71811A92E001F1811802600271812F82C207F97840607080093DFC142201B82A1009AA0A01E428027D012C678B00E78B666491646580897A007A00658064907C80383A6465816503E5FFE4E83BC00C646582AC678B28027D0109E5B589666664B8FD80400FE3603FA00FA40F82854120870542013541403C85004FA0258CF1601CF16CCC922C8CB0112F400F400CB00C9F9007074C8CB02CA07CBFFC9D05008C705F2E04A12A1035024C85004FA0258CF16CCCCC9ED5401FA403020D70B01C3008E1F8210D53276DB708010C8CB055003CF1622FA0212CB6ACB1FCB3FC98042FB00915BE200303515C705F2E049FA403059C85004FA0258CF16CCCCC9ED54002E5143C705F2E049D43001C85004FA0258CF16CCCCC9ED54007DADBCF6A2687D007D206A6A183618FC1400B82A1009AA0A01E428027D012C678B00E78B666491646580897A007A00658064FC80383A6465816503E5FFE4E840001FAF16F6A2687D007D206A6A183FAA904051007F09'

    def __init__(self, **kwargs):
        self.code = kwargs.get('code') or self.code
        kwargs['code'] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)

    def create_data_cell(self) -> Cell:
        cell = Cell()
        cell.bits.write_grams(0)  # total supply
        cell.bits.write_address(self.options['admin_address'])
        cell.refs.append(create_offchain_uri_cell(self.options['jetton_content_uri']))
        cell.refs.append(Cell.one_from_boc(self.options['jetton_wallet_code_hex']))
        return cell

    def create_mint_body(self, destination: Address, jetton_amount: int, amount: int = 50000000, query_id: int = 0) -> Cell:
        body = Cell()
        body.bits.write_uint(21, 32)  # OP mint
        body.bits.write_uint(query_id, 64)
        body.bits.write_address(destination)
        body.bits.write_grams(amount)

        transfer_body = Cell()  # internal transfer
        transfer_body.bits.write_uint(0x178d4519, 32)  # OP transfer
        transfer_body.bits.write_uint(query_id, 64)
        transfer_body.bits.write_grams(jetton_amount)  # jetton amount
        transfer_body.bits.write_address(None)  # from_address
        transfer_body.bits.write_address(None)  # response_address
        transfer_body.bits.write_grams(0)  # forward amount
        transfer_body.bits.write_bit(0)  # forward_payload in this slice, not separate cell

        body.refs.append(transfer_body)
        return body

    def create_change_admin_body(self, new_admin_address: Address, query_id: int = 0) -> Cell:
        body = Cell()
        body.bits.write_uint(3, 32)  # OP
        body.bits.write_uint(query_id, 64)  # query_id
        body.bits.write_address(new_admin_address)
        return body

    def create_edit_content_body(self, jetton_content_uri: str, query_id: int = 0) -> Cell:
        body = Cell()
        body.bits.write_uint(4, 32)  # OP
        body.bits.write_uint(query_id, 64)  # query_id
        body.refs.append(create_offchain_uri_cell(jetton_content_uri))
        return body




import codecs
import ctypes
import math
import struct

import nacl
from nacl.bindings import crypto_sign, crypto_sign_BYTES
from nacl.signing import SignedMessage


def concat_bytes(a, b):
    return a + b  # ?


def move_to_end(index_hashmap, topological_order_arr, target):
    target_index = index_hashmap[target]
    for _hash in index_hashmap:
        if index_hashmap[_hash] > target_index:
            index_hashmap[_hash] -= 1
    index_hashmap[target] = len(topological_order_arr) - 1
    data = topological_order_arr.pop(target_index)
    topological_order_arr.append(data)
    for sub_cell in data[1].refs:
        topological_order_arr, index_hashmap = move_to_end(index_hashmap, topological_order_arr, sub_cell.bytes_hash())
    return [topological_order_arr, index_hashmap]


def tree_walk(cell, topological_order_arr, index_hashmap, parent_hash=None):
    cell_hash = cell.bytes_hash()
    if cell_hash in index_hashmap:
        if parent_hash:
            if index_hashmap[parent_hash] > index_hashmap[cell_hash]:
                topological_order_arr, index_hashmap = move_to_end(index_hashmap, topological_order_arr, cell_hash)
        return [topological_order_arr, index_hashmap]

    index_hashmap[cell_hash] = len(topological_order_arr)
    topological_order_arr.append([cell_hash, cell])
    for sub_cell in cell.refs:
        topological_order_arr, index_hashmap = tree_walk(sub_cell, topological_order_arr, index_hashmap, cell_hash)
    return [topological_order_arr, index_hashmap]


def _crc32c(crc, bytes_arr):
    POLY = 0x82f63b78

    crc ^= 0xffffffff

    for n in range(len(bytes_arr)):
        crc ^= bytes_arr[n]
        crc = (crc >> 1) ^ POLY if crc & 1 else crc >> 1
        crc = (crc >> 1) ^ POLY if crc & 1 else crc >> 1
        crc = (crc >> 1) ^ POLY if crc & 1 else crc >> 1
        crc = (crc >> 1) ^ POLY if crc & 1 else crc >> 1
        crc = (crc >> 1) ^ POLY if crc & 1 else crc >> 1
        crc = (crc >> 1) ^ POLY if crc & 1 else crc >> 1
        crc = (crc >> 1) ^ POLY if crc & 1 else crc >> 1
        crc = (crc >> 1) ^ POLY if crc & 1 else crc >> 1

    return crc ^ 0xffffffff


def crc32c(bytes_array):
    int_crc = _crc32c(0, bytes_array)

    # TODO: check mistakes
    arr = bytearray(4)
    struct.pack_into('>I', arr, 0, int_crc)

    return bytes(arr)[::-1]


def crc16(data):
    POLY = 0x1021
    reg = 0
    message = bytes(data) + bytes(2)

    for byte in message:
        mask = 0x80
        while mask > 0:
            reg <<= 1
            if byte & mask:
                reg += 1
            mask >>= 1
            if reg > 0xffff:
                reg &= 0xffff
                reg ^= POLY

    return bytearray([math.floor(reg / 256), reg % 256])


def read_n_bytes_uint_from_array(size_bytes, uint8_array):
    res = 0
    for c in range(size_bytes):
        res *= 256
        res += uint8_array[c]  # must be uint8

    return res


def compare_bytes(bytes_1, bytes_2):
    return str(bytes_1) == str(bytes_2)  # why str?


def string_to_bytes(string, size=1):  # ?
    if size == 1:
        buf = (ctypes.c_uint8 * len(string))()
    elif size == 2:
        buf = (ctypes.c_uint16 * len(string) * 2)()
    elif size == 4:
        buf = (ctypes.c_uint32 * len(string) * 4)()

    for i, c in enumerate(string):
        # buf[i] = ord(c)
        buf[i] = c  # ?

    return bytes(buf)


def sign_message(message: bytes,
                 signing_key,
                 encoder: nacl.encoding.Encoder = nacl.encoding.RawEncoder, ) -> SignedMessage:
    raw_signed = crypto_sign(message, signing_key)

    signature = encoder.encode(raw_signed[:crypto_sign_BYTES])
    message = encoder.encode(raw_signed[crypto_sign_BYTES:])
    signed = encoder.encode(raw_signed)

    return SignedMessage._from_parts(signature, message, signed)


def b64str_to_bytes(b64str):
    b64bytes = codecs.encode(b64str, "utf-8")
    return codecs.decode(b64bytes, "base64")


def b64str_to_hex(b64str):
    _bytes = b64str_to_bytes(b64str)
    _hex = codecs.encode(_bytes, "hex")
    return codecs.decode(_hex, "utf-8")


def bytes_to_b64str(bytes_arr):
    return codecs.decode(codecs.encode(
        bytes_arr, "base64"), 'utf-8').replace("\n", '')



import asyncio

from conftest import vcr_c

from aiotx.clients import AioTxTONClient


@vcr_c.use_cassette("tests/fixtures/cassettes/ton/test_async_monitoring.yaml")
async def test_async_monitoring(ton_mainnet_client: AioTxTONClient):
    blocks = []
    transactions = []

    @ton_mainnet_client.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @ton_mainnet_client.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await ton_mainnet_client.start_monitoring(38093046)
    try:
        await asyncio.sleep(3)
    except KeyboardInterrupt:
        ton_mainnet_client.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    assert 38093046 in blocks
    assert 38093047 in blocks
    assert 38093048 in blocks

    # transactions from shard 43736268 of master block 38093046
    # https://tonscan.org/block/0:6000000000000000:43736268
    assert "r4r7FlGRAflxcnfEeDwVy1uqhEh/ajyZsUCgrwbATTY=" in [tx["hash"] for tx in transactions]
    # transactions from shard 43736268 of master block 38093046
    assert "NZEK2M1V3VqEdhuEmj5qayWh2e0Sc/lEcOYGMpUDe0o=" in [tx["hash"] for tx in transactions]

    # transactions from shard 43735772 of master block 38093046
    assert "ayDy6PrzGMCR4rMMD3FEThLEgjMB8heqmcUdJC3CI/A=" in [tx["hash"] for tx in transactions]
    # transactions from shard 43735772 of master block 38093046
    assert "EPfqGpYOWZg1LpGQ06HfxFhxwwWNZKeQQtIAJ4U60Eg=" in [tx["hash"] for tx in transactions]

    # transactions from shard 43736270 of master block 38093047
    assert "xcPkFFntBatwm/mC06Enl3tYZmCQhnfcQncZ0InbA3o=" in [tx["hash"] for tx in transactions]

    # transactions from shard 43736027 of master block 38093047
    assert "WoQxVXUxtxC8nPdkvPdj0/wVTb6JkntyXg1fXxmdXQQ=" in [tx["hash"] for tx in transactions]
    assert "/NiBkmuPCaFdu4pv+Y1oxdJ0dlOWkEgCFPf+Mg/S6LU=" in [tx["hash"] for tx in transactions]
    assert "GKZ+EwkxxAONtWntnUTD/ISg2aZ2iy16XkVS/wiS3Lw=" in [tx["hash"] for tx in transactions]


@vcr_c.use_cassette("tests/fixtures/cassettes/ton/test_async_monitoring_testnet.yaml")
async def test_async_monitoring_testnet(ton_client: AioTxTONClient):
    blocks = []
    transactions = []

    @ton_client.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @ton_client.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await ton_client.start_monitoring(19627950)
    try:
        await asyncio.sleep(3)
    except KeyboardInterrupt:
        ton_client.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    assert 19627950 in blocks
    assert 19627951 in blocks
    assert 19627952 in blocks

    # we are not able to see that in block explorer tonscan, but we can see that in ton centre
    assert "FT1lZcFBL2rR3KYtz6m9nWTegS76u4kudfr+4stIxoU=" in [tx["hash"] for tx in transactions]
    assert "jxP0qffivNEbR1SVgG2jZ5ds+Hk/aUKW0rmY0/Jso9Q=" in [tx["hash"] for tx in transactions]



import asyncio
import logging
import os

import pytest
import vcr
from pytest import FixtureRequest

from aiotx.clients import (
    AioTxBSCClient,
    AioTxBTCClient,
    AioTxETHClient,
    AioTxLTCClient,
    AioTxPolygonClient,
    AioTxTONClient,
    AioTxTRONClient,
)
from aiotx.log import set_logger_level

set_logger_level("INFO")

# ALL = "all"
# ANY = "any"
# NEW_EPISODES = "new_episodes"
# NONE = "none"
# ONCE = "once"

BSC_TEST_NODE_URL = "https://bsc-testnet-rpc.publicnode.com"
ETH_TEST_NODE_URL = "https://ethereum-sepolia-rpc.publicnode.com"
LTC_TEST_NODE_URL = "https://api.tatum.io/v3/blockchain/node/litecoin-core-testnet"
BTC_TEST_NODE_URL = "https://bitter-warmhearted-film.btc-testnet.quiknode.pro/ddf238a16e9d7342e403d000a6134a2d750765f8/"
TON_MAINNET_NODE_URL = "https://go.getblock.io/875fb0dee2544bb0bc59dd08c6f39330/jsonRPC"
POLYGON_TEST_NODE_URL = "https://polygon-amoy-bor-rpc.publicnode.com"
TON_TEST_NODE_URL = "https://testnet.toncenter.com/api/v2/jsonRPC"
TRON_TEST_NODE_URL = "https://go.getblock.io/d5d80fb886f2401a98a7f4fe57d88839"

pytest.mark.mysql = pytest.mark.mysql
logging.getLogger("vcr").setLevel(logging.WARNING)
vcr_c = vcr.VCR(
    cassette_library_dir="tests/fixtures/cassettes",
    record_mode=os.environ.get("VCR_RECORD_MODE", "none"),
    match_on=["host", "path", "method", "query", "raw_body", "body"],
    filter_headers=["Authorization", "Cookie", "Date"],
)

@pytest.fixture
def ton_client() -> AioTxTONClient:
    # current test rpc connection returning -1 as workchain but it should be 0, 
    # so we are setting that param by ourself
    return AioTxTONClient(TON_TEST_NODE_URL, workchain=0)

@pytest.fixture
def ton_mainnet_client() -> AioTxTONClient:
    # testnet block monitoring is not working as 
    # expected, and we using mainnet to test monitoring
    return AioTxTONClient(TON_MAINNET_NODE_URL)

@pytest.fixture
def tron_client() -> AioTxTRONClient:
    return AioTxTRONClient(TRON_TEST_NODE_URL)

@pytest.fixture
def bsc_client() -> AioTxBSCClient:
    return AioTxBSCClient(BSC_TEST_NODE_URL)

@pytest.fixture
def polygon_client() -> AioTxPolygonClient:
    return AioTxPolygonClient(POLYGON_TEST_NODE_URL)

@pytest.fixture
def eth_client() -> AioTxETHClient:
    return AioTxETHClient(ETH_TEST_NODE_URL)


@pytest.fixture
def ltc_public_client(request: FixtureRequest) -> AioTxLTCClient:
    def teardown():
        try:
            os.remove("test_ltc.sqlite")
        except FileNotFoundError:
            print("test_ltc.sqlite FileNotFoundError")

    request.addfinalizer(teardown)
    return AioTxLTCClient(LTC_TEST_NODE_URL, testnet=True, db_url="sqlite+aiosqlite:///test_ltc.sqlite")


@pytest.fixture
def btc_client(request: FixtureRequest) -> AioTxBTCClient:
    def teardown():
        try:
            os.remove("test_btc.sqlite")
        except FileNotFoundError:
            print("test_btc.sqlite FileNotFoundError")

    request.addfinalizer(teardown)
    return AioTxBTCClient(BTC_TEST_NODE_URL, testnet=True, db_url="sqlite+aiosqlite:///test_btc.sqlite")


@pytest.fixture
def btc_client_mysql(request: FixtureRequest) -> AioTxBTCClient:
    aiotx_btc_mysql_client = AioTxBTCClient(
        BTC_TEST_NODE_URL,
        testnet=True,
        db_url=f"mysql+aiomysql://{os.getenv('DB_USER')}:{os.getenv('DB_PASSWORD')}@{os.getenv('DB_HOST')}:{os.getenv('DB_PORT')}/{os.getenv('DB_NAME')}",
    )


    def teardown():
        asyncio.run(aiotx_btc_mysql_client.monitor._drop_tables())

    request.addfinalizer(teardown)
    return aiotx_btc_mysql_client

@pytest.fixture
def ltc_client_mysql(request: FixtureRequest) -> AioTxLTCClient:
    aiotx_ltc_mysql_client = AioTxLTCClient(
        LTC_TEST_NODE_URL,
        testnet=True,
        db_url=f"mysql+aiomysql://{os.getenv('DB_USER')}:{os.getenv('DB_PASSWORD')}@{os.getenv('DB_HOST')}:{os.getenv('DB_PORT')}/{os.getenv('DB_NAME')}",
    )


    def teardown():
        asyncio.run(aiotx_ltc_mysql_client.monitor._drop_tables())

    request.addfinalizer(teardown)
    return aiotx_ltc_mysql_client


import decimal
import json
from typing import Optional, Union

import aiohttp

from aiotx.clients._base_client import AioTxClient, BlockMonitor
from aiotx.exceptions import (
    BlockNotFoundError,
    InvalidArgumentError,
    RpcConnectionError,
    WrongPrivateKey,
)
from aiotx.utils.tonsdk.contract.wallet import Wallets, WalletVersionEnum
from aiotx.utils.tonsdk.crypto import mnemonic_new
from aiotx.utils.tonsdk.crypto._mnemonic import mnemonic_is_valid
from aiotx.utils.tonsdk.utils import bytes_to_b64str
from aiotx.utils.tonsdk.utils import from_nano as tonsdk_from_nano
from aiotx.utils.tonsdk.utils import to_nano as tonsdk_to_nano


class AioTxTONClient(AioTxClient):
    def __init__(
        self, node_url, wallet_version: WalletVersionEnum = WalletVersionEnum.v4r2, workchain: Optional[int] = None
    ):
        super().__init__(node_url)
        self.monitor = TonMonitor(self)
        self._monitoring_task = None
        self.workchain = workchain
        self.wallet_version = wallet_version

    async def generate_address(self) -> tuple[str, str, str]:
        if self.workchain is None:
            await self._get_network_params()
        wallet_mnemonics = mnemonic_new()
        _mnemonics, _, _, wallet = Wallets.from_mnemonics(wallet_mnemonics, self.wallet_version, self.workchain)
        return (
            " ".join(_mnemonics),
            wallet.address.to_string(is_user_friendly=True, is_url_safe=True),
            wallet.address.to_string(False, False, False),
        )

    def _unpack_mnemonic(self, mnemonic_str: str):
        assert isinstance(mnemonic_str, str), "Mnemonic should be represented as string!"
        mnemonic_list = mnemonic_str.split(" ")
        if not mnemonic_is_valid(mnemonic_list):
            raise WrongPrivateKey("mnemonic phrase not valid!")
        return mnemonic_list

    async def get_address_from_mnemonics(self, wallet_mnemonic: str) -> tuple[str, str]:
        if self.workchain is None:
            await self._get_network_params()
        wallet_mnemonic_list = self._unpack_mnemonic(wallet_mnemonic)
        _, _, _, wallet = Wallets.from_mnemonics(wallet_mnemonic_list, self.wallet_version, self.workchain)
        return wallet.address.to_string(is_user_friendly=True, is_url_safe=True), wallet.address.to_string(
            False, False, False
        )
    
    async def get_last_master_block(self) -> int:
        payload = {"method": "getMasterchainInfo", "params": {}}
        result = await self._make_rpc_call(payload)
        last_block = result["last"]
        return last_block

    async def get_master_block_shards(self, seqno: int) -> list[dict]:
        payload = {"method": "shards", "params": {"seqno": seqno}}
        result = await self._make_rpc_call(payload)
        return result["shards"]

    async def get_balance(self, address) -> int:
        payload = {"method": "getAddressBalance", "params": {"address": address}}
        balance_info = await self._make_rpc_call(payload)
        return int(balance_info)

    async def get_address_information(self, address):
        payload = {"method": "getAddressInformation", "params": {"address": address}}
        information = await self._make_rpc_call(payload)
        return information

    async def get_wallet_information(self, address):
        payload = {"method": "getWalletInformation", "params": {"address": address}}
        information = await self._make_rpc_call(payload)
        return information

    async def get_transaction_count(self, address) -> int:
        address_information = await self.get_wallet_information(address)
        return address_information.get("seqno", 0)

    async def pack_address(self, address) -> str:
        payload = {"method": "packAddress", "params": {"address": address}}
        packed_address = await self._make_rpc_call(payload)
        return packed_address

    async def get_block_transactions(self, workchain, shard, seqno, count=40) -> list[dict]:
        # Here we don't have workchain by default, because in mainnet
        # for example master block workchain is -1 and shard is 0
        # So we should use workchain here
        payload = {
            "method": "getBlockTransactions",
            "params": {"workchain": workchain, "shard": shard, "seqno": seqno, "count": count},
        }
        information = await self._make_rpc_call(payload)
        return information["transactions"]

    async def detect_address(self, address) -> dict:
        payload = {"method": "detectAddress", "params": {"address": address}}
        information = await self._make_rpc_call(payload)
        return information
    
    def to_nano(self, number: Union[int, float, str, decimal.Decimal], unit: str = "ton")-> int:
        return tonsdk_to_nano(number, unit)
    
    def from_nano(self, number: int, unit: str = "ton")-> int:
        return tonsdk_from_nano(number, unit)

    async def send(self, mnemonic: str, to_address: str, amount: int, seqno: int = None, memo: str = None) -> str:
        assert isinstance(amount, int), "Amount should be integer! Please use to_nano for convert it!"
        if self.workchain is None:
            await self._get_network_params()

        from_address, _ = await self.get_address_from_mnemonics(mnemonic)
        if seqno is None:
            seqno = await self.get_transaction_count(from_address)
        
        boc = self._create_transfer_boc(mnemonic, to_address, amount, seqno, memo)
        boc_answer_data = await self.send_boc_return_hash(boc)
        return boc_answer_data["hash"]
    
    # async def estimate_fee(self, boc: str, address: str):
    #     payload = {"method": "estimateFee", "params": {"address": address,"body": boc}}
    #     information = await self._make_rpc_call(payload)
    #     return information

    def _create_transfer_boc(self, mnemonic_str: str, to_address, amount, seqno, memo) -> str:
        mnemonic_list = self._unpack_mnemonic(mnemonic_str)
        _, _, _, wallet = Wallets.from_mnemonics(mnemonic_list, self.wallet_version, self.workchain)
        query = wallet.create_transfer_message(
            to_addr=to_address, amount=amount, payload=memo, seqno=seqno
        )
        boc = bytes_to_b64str(query["message"].to_boc(False))
        return boc

    async def send_boc_return_hash(self, boc) -> dict:
        payload = {"method": "sendBocReturnHash", "params": {"boc": boc}}
        information = await self._make_rpc_call(payload)
        return information

    async def _get_network_params(self) -> tuple[int, int, int]:
        master_block_data = await self.get_last_master_block()
        workchain = master_block_data["workchain"]
        shard = master_block_data["shard"]
        seqno = master_block_data["seqno"]
        self.workchain = workchain
        return workchain, shard, seqno

    async def get_transactions(
        self, address, limit: int = None, lt: int = None, hash: str = None, to_lt: int = None, archival: bool = None
    ) -> list[dict]:
        """
        Retrieves transactions for a given TON account.

        Args:
            address (str): Identifier of the target TON account in any form.
            limit (int, optional): Maximum number of transactions in the response.
            lt (int, optional): Logical time of the transaction to start with. Must be sent with `hash`.
            hash (str, optional): Hash of the transaction to start with, in base64 or hex encoding. Must be sent with `lt`.
            to_lt (int, optional): Logical time of the transaction to finish with (to get transactions from `lt` to `to_lt`).
            archival (bool, optional): By default, the `getTransaction` request is processed by any available liteserver.
                If `archival=True`, only liteservers with full history are used.
        """
        params = {"address": address}
        if lt is not None:
            params["lt"] = lt
        if hash is not None:
            params["hash"] = hash
        if to_lt is not None:
            params["to_lt"] = to_lt
        if archival is not None:
            params["archival"] = archival
        if limit is not None:
            params["limit"] = limit

        payload = {"method": "getTransactions", "params": params}
        information = await self._make_rpc_call(payload)
        return information

    async def _make_rpc_call(self, payload) -> dict:
        payload["jsonrpc"] = "2.0"
        payload["id"] = 1
        payload_json = json.dumps(payload)
        headers = {"Content-Type": "application/json"}
        async with aiohttp.ClientSession() as session:
            async with session.post(self.node_url, data=payload_json, headers=headers) as response:
                response_text = await response.text()
                if response.status != 200:
                    if "cannot find block" in response_text:
                        raise BlockNotFoundError(response_text)
                    if "Incorrect address" in response_text:
                        raise InvalidArgumentError(response_text)
                    raise RpcConnectionError(f"Node response status code: {response.status} response test: {response_text}")
                result = await response.json()
                return result["result"]


class TonMonitor(BlockMonitor):
    def __init__(self, client: AioTxTONClient, last_master_block: Optional[int] = None):
        self.client = client
        self.block_handlers = []
        self.transaction_handlers = []
        self.running = False
        self._last_master_block = last_master_block

    async def poll_blocks(
        self,
    ):
        workchain, shard, seqno = await self.client._get_network_params()
        if self.client.workchain is None:
            self.client.workchain = workchain
        target_block = seqno if self._latest_block is None else self._latest_block
        if target_block > seqno:
            return
        shards = await self.client.get_master_block_shards(target_block)
        for shard in shards:
            shard_transactions = await self.client.get_block_transactions(
                shard["workchain"], shard["shard"], shard["seqno"], 1000
            )
            await self.process_shard_transactions(shard_transactions)
        await self.process_master_block(target_block)
        self._latest_block = target_block + 1

    async def process_master_block(self, block):
        for handler in self.block_handlers:
            await handler(block)

    async def process_shard_transactions(self, shard_transactions):
        for transaction in shard_transactions:
            for handler in self.transaction_handlers:
                await handler(transaction)




class AioTxError(Exception):
    """
    Base exception for all aiotx errors.
    """


class BlockNotFoundError(AioTxError):
    pass


class InternalJSONRPCError(AioTxError):
    pass


class StackLimitReachedError(AioTxError):
    pass


class MethodHandlerCrashedError(AioTxError):
    pass


class ExecutionTimeoutError(AioTxError):
    pass


class NonceTooLowError(AioTxError):
    pass


class FilterNotFoundError(AioTxError):
    pass


class TraceRequestLimitExceededError(AioTxError):
    pass


class TransactionCostExceedsGasLimitError(AioTxError):
    pass


class NetworkError(AioTxError):
    pass


class VMExecutionError(AioTxError):
    pass


class MethodNotFoundError(AioTxError):
    pass


class InvalidRequestError(AioTxError):
    pass


class InvalidArgumentError(AioTxError):
    pass


class BlockRangeLimitExceededError(AioTxError):
    pass


class InternalJSONAioTxError(AioTxError):
    pass


class TransactionNotFound(AioTxError):
    pass


class ReplacementTransactionUnderpriced(AioTxError):
    pass


class WrongPrivateKey(AioTxError):
    pass


class InsufficientFunds(AioTxError):
    pass

class RpcConnectionError(AioTxError):
    pass



def find_common_prefix(src):
    # Corner cases
    if len(src) == 0:
        return ''
    if len(src) == 1:
        return src[0]

    # Searching for prefix
    _sorted = sorted(src)
    size = 0
    for i, e in enumerate(_sorted[0]):
        if e == _sorted[-1][i]:
            size += 1
        else:
            break

    return _sorted[0][:size]




import asyncio


class AioTxClient:
    def __init__(self, node_url):
        self.node_url = node_url
        self.monitor = BlockMonitor(self)
        self._monitoring_task = None

    async def start_monitoring(self, monitoring_start_block: int = None):
        if self._monitoring_task is None:
            self._monitoring_task = asyncio.create_task(self._run_monitoring(monitoring_start_block))
        return self._monitoring_task

    def stop_monitoring(self):
        if self._monitoring_task is not None:
            self._monitoring_task.cancel()
            try:
                asyncio.get_event_loop().run_until_complete(self._monitoring_task)
            except asyncio.CancelledError:
                pass
            self._monitoring_task = None

    async def _run_monitoring(self, monitoring_start_block):
        try:
            async with self.monitor:
                await self.monitor.start(monitoring_start_block)
        except asyncio.CancelledError:
            pass


class BlockMonitor:
    def __init__(self, client: AioTxClient):
        self.client = client
        self.block_handlers = []
        self.transaction_handlers = []
        self.running = False
        self._latest_block = None

    def on_block(self, func):
        self.block_handlers.append(func)
        return func

    def on_transaction(self, func):
        self.transaction_handlers.append(func)
        return func

    async def __aenter__(self):
        self.running = True
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    async def start(self, monitoring_start_block):
        self.running = True
        self._latest_block = monitoring_start_block
        while self.running:
            try:
                await self.poll_blocks()
                await asyncio.sleep(1)
            except Exception as e:
                print(f"Error during polling: {e}")
                await asyncio.sleep(2)

    def stop(self):
        self.running = False

    async def poll_blocks(self):
        pass

    async def process_block(self, block):
        pass



import asyncio

from conftest import vcr_c

from aiotx.clients import AioTxPolygonClient


@vcr_c.use_cassette("tests/fixtures/cassettes/polygon/test_async_monitoring.yaml")
async def test_async_monitoring(polygon_client: AioTxPolygonClient):
    blocks = []
    transactions = []

    @polygon_client.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @polygon_client.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await polygon_client.start_monitoring(7385898)
    try:
        await asyncio.sleep(3)
    except KeyboardInterrupt:
        polygon_client.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    assert 7385898 in blocks
    assert 7385899 in blocks
    assert "0x568435d3c40e00160885d9030f231adfc230c91a5dab740515e89e30600a734c" in [tx["hash"] for tx in transactions]
    assert "0x1913ae3314b38254c07538910ebdf2e3a3d6562b084769ef6614a0f0f4e01367" in [tx["hash"] for tx in transactions]

    for tx in transactions:
        assert "aiotx_decoded_input" in tx.keys()



from ._address import Address
from ._currency import TonCurrencyEnum, from_nano, to_nano
from ._exceptions import InvalidAddressError
from ._utils import (
    b64str_to_bytes,
    b64str_to_hex,
    bytes_to_b64str,
    compare_bytes,
    concat_bytes,
    crc16,
    crc32c,
    move_to_end,
    read_n_bytes_uint_from_array,
    sign_message,
    tree_walk,
)

__all__ = [
    'Address',
    'InvalidAddressError',

    'concat_bytes',
    'move_to_end',
    'tree_walk',
    'crc32c',
    'crc16',
    'read_n_bytes_uint_from_array',
    'compare_bytes',
    'sign_message',
    'b64str_to_bytes',
    'b64str_to_hex',
    'bytes_to_b64str',

    'to_nano',
    'from_nano',
    'TonCurrencyEnum',
]



import asyncio

from conftest import vcr_c

from aiotx.clients import AioTxBSCClient


@vcr_c.use_cassette("tests/fixtures/cassettes/tron/test_async_monitoring.yaml")
async def test_async_monitoring(tron_client: AioTxBSCClient):
    blocks = []
    transactions = []

    @tron_client.monitor.on_block
    async def handle_block(block):
        blocks.append(block)

    @tron_client.monitor.on_transaction
    async def handle_transaction(transaction):
        transactions.append(transaction)

    await tron_client.start_monitoring(47338868)
    try:
        await asyncio.sleep(3)
    except KeyboardInterrupt:
        tron_client.stop_monitoring()

    assert len(blocks) > 0
    assert len(transactions) > 0

    assert 47338868 in blocks
    assert 47338869 in blocks
    assert 47338870 in blocks

    # TRX transfer from block 47338868
    assert "0x72ce7a6a1bfdaca67431e96857cb17dc1c86bbe36cfefb4bab7ed72f7e72122c" in [tx["hash"] for tx in transactions]

    #TRC20 transaction from block 47338870
    assert "0x2ee04f0cc089e1f1d2499c1f7ee3114a69a61dda478c39ea2d1becdd97d387fa" in [tx["hash"] for tx in transactions]
    assert "0xc00858d686c1433369815ba366610b84064cbc4a684284887ccbf2091ed80558" in [tx["hash"] for tx in transactions]
    
    for tx in transactions:
        assert "aiotx_decoded_input" in tx.keys()



from ....boc import Cell
from ....utils import Address
from ... import Contract


class NFTItem(Contract):
    code = 'B5EE9C7241020D010001D0000114FF00F4A413F4BCF2C80B0102016202030202CE04050009A11F9FE00502012006070201200B0C02D70C8871C02497C0F83434C0C05C6C2497C0F83E903E900C7E800C5C75C87E800C7E800C3C00812CE3850C1B088D148CB1C17CB865407E90350C0408FC00F801B4C7F4CFE08417F30F45148C2EA3A1CC840DD78C9004F80C0D0D0D4D60840BF2C9A884AEB8C097C12103FCBC20080900113E910C1C2EBCB8536001F65135C705F2E191FA4021F001FA40D20031FA00820AFAF0801BA121945315A0A1DE22D70B01C300209206A19136E220C2FFF2E192218E3E821005138D91C85009CF16500BCF16712449145446A0708010C8CB055007CF165005FA0215CB6A12CB1FCB3F226EB39458CF17019132E201C901FB00104794102A375BE20A00727082108B77173505C8CBFF5004CF1610248040708010C8CB055007CF165005FA0215CB6A12CB1FCB3F226EB39458CF17019132E201C901FB000082028E3526F0018210D53276DB103744006D71708010C8CB055007CF165005FA0215CB6A12CB1FCB3F226EB39458CF17019132E201C901FB0093303234E25502F003003B3B513434CFFE900835D27080269FC07E90350C04090408F80C1C165B5B60001D00F232CFD633C58073C5B3327B5520BF75041B'

    def __init__(self, **kwargs):
        self.code = kwargs.get('code') or self.code
        kwargs["code"] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)

    def create_data_cell(self) -> Cell:
        cell = Cell()
        cell.bits.write_uint(self.options.get('index', 0), 64)
        cell.bits.write_address(self.options.get('collection_address', None))
        if 'owner_address' in self.options:
            cell.bits.write_address(self.options['owner_address'])
        if 'content' in self.options:
            cell.refs.append(self.options['content'])
        return cell

    def create_transfer_body(
            self, new_owner_address: Address, response_address: Address = None,
            forward_amount: int = 0, forward_payload: bytes = None, query_id: int = 0
    ) -> Cell:
        cell = Cell()
        cell.bits.write_uint(0x5fcc3d14, 32)  # transfer OP
        cell.bits.write_uint(query_id, 64)
        cell.bits.write_address(new_owner_address)
        cell.bits.write_address(response_address or new_owner_address)
        cell.bits.write_bit(False)  # null custom_payload
        cell.bits.write_grams(forward_amount)
        cell.bits.write_bit(False)  # forward_payload in this slice, not separate cell
        if forward_payload:
            cell.bits.write_bytes(forward_payload)

        return cell

    def create_get_static_data_body(self, query_id: int = 0) -> Cell:
        cell = Cell()
        cell.bits.write_uint(0x2fcb26a2, 32)
        cell.bits.write_uint(query_id, 64)
        return cell






import pytest
from conftest import vcr_c

from aiotx.clients import AioTxTRONClient
from aiotx.exceptions import (
    InvalidArgumentError,
    RpcConnectionError,
    TransactionNotFound,
)

CONTRACT = "TXLAQ63Xg1NAzckPwKHvzw7CSEmLMEqcdj"

async def test_wallet_generator_convertor(tron_client: AioTxTRONClient):
    wallet = tron_client.generate_address()
    print("wallet", wallet)

    base_58_from_hex = tron_client.hex_address_to_base58(wallet["hex_address"])
    hex_from_base58 = tron_client.base58_to_hex_address(wallet["base58check_address"])

    wallet_from_key = tron_client.get_address_from_private_key(wallet["private_key"])

    assert wallet == wallet_from_key
    assert base_58_from_hex == wallet["base58check_address"]
    assert hex_from_base58 == wallet["hex_address"]


@vcr_c.use_cassette("tron/get_last_block.yaml")
async def test_get_last_block(tron_client: AioTxTRONClient):
    block_id = await tron_client.get_last_block_number()
    assert isinstance(block_id, int)


@vcr_c.use_cassette("tron/get_chain_id.yaml")
async def test_get_chain_id(tron_client: AioTxTRONClient):
    chain_id = await tron_client.get_chain_id()
    assert isinstance(chain_id, int)
    assert chain_id == 3448148188


@pytest.mark.parametrize(
    "wallet_address, expected_exception, expected_balance",
    [
        ("TEZQQ5BXq3nFKUFJknoV15CW24twzH81La", None, 4001000000),
        ("TCnvuKGHUjjx6z4YAoUtCopXA8pXzv16YZ", None, 2892750560),
        ("TX9ctsGJSx8LniV1JgPuvNkXdTGrDdyZqt", None, 390613840),
        ("TXLnrKzcnTqPDiRYgECNMCC37wKgMvZkKU", None, 61477500),
        ("TGYycKuwtyJ1zRt196PhaWfPuUPGr1wi2P", None, 23520),
        ("TAhMH42Mr61MRLkbEpT1vjbXpsKSXBtHRz", None, 0),
        ("TWkcsRj1FnAXA1HzWZEW93hdoQxQ2YXFN", ValueError, 0),
    ],
)
@vcr_c.use_cassette("tron/get_balance.yaml")
async def test_get_balance(tron_client: AioTxTRONClient, wallet_address, expected_exception, expected_balance):
    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.get_balance(wallet_address)
    else:
        balance = await tron_client.get_balance(wallet_address)
        assert isinstance(balance, int)
        assert balance == expected_balance


@pytest.mark.parametrize(
    "tx_id, expected_exception",
    [
        ("a7969308be486dcbd5bc16f9a9e750249f1ce4ff5ad36693cd56a8aa81159df8", None),
        ("83033077a7a5ea6e07fa1e36069886daf3abe8bae4e12168391db9c0b28002a7", None),
        ("28777a1b17916030c99a3da3b88e399c11abe3cbbc9e1319e2d2edd83d2810c8", None),
        ("301684ded91391888927f4c3761af0ad76659ede03033fd12c20754e6bc69821", None),
        ("603d90143f388737fa0c57976650301fb493ef19c513c9076099335d0038f2ee", None),
        ("0820f27e5a67ee3443d0b4a60a1f6c6625328b074dedc2f17c0ea8c44ea440e5", None),
        ("0820f27e5a67ee3443d0b4a60a1f6c6625328b074dedc2f17c0ea8c44ea440e6", TransactionNotFound),
        ("0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf2701b2", InvalidArgumentError),
        ("0x93ff17f18511c4fa74166a2dbd53ffa7ce5cc2ee431cbb4fd5e02bbdf701b72", InvalidArgumentError),
    ],
)
@vcr_c.use_cassette("tron/get_transaction.yaml")
async def test_get_transaction(tron_client: AioTxTRONClient, tx_id, expected_exception):
    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.get_transaction(tx_id)
    else:
        tx = await tron_client.get_transaction(tx_id)
        assert "aiotx_decoded_input" in tx.keys()


@pytest.mark.parametrize(
    "wallet_address, contract, expected_exception, expected_balance",
    [
        ("TEZQQ5BXq3nFKUFJknoV15CW24twzH81La", CONTRACT, None, 50000000000),
        ("TCnvuKGHUjjx6z4YAoUtCopXA8pXzv16YZ", CONTRACT, None, 163489214426),
        ("TX9ctsGJSx8LniV1JgPuvNkXdTGrDdyZqt", CONTRACT, None, 0),
        ("TXLnrKzcnTqPDiRYgECNMCC37wKgMvZkKU", CONTRACT, None, 49845248),
        ("TGYycKuwtyJ1zRt196PhaWfPuUPGr1wi2P", CONTRACT, None, 0),
        ("TGYycKuwtyJ1zRt196PhaWfPuUPGr1wi2D", CONTRACT, ValueError, 0),
        ("TAhMH42Mr61MRLkbEpT1vjbXpsKSXBtHRz", CONTRACT, None, 0),
        ("0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C", "Eeaa7E347ee4557f9a6F71C", InvalidArgumentError, 0),
    ],
)
@vcr_c.use_cassette("tron/get_token_balance.yaml")
async def test_get_token_balance(
    tron_client: AioTxTRONClient, wallet_address, contract, expected_exception, expected_balance
):

    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.get_contract_balance(wallet_address, contract)
    else:
        balance = await tron_client.get_contract_balance(wallet_address, contract)
        assert isinstance(balance, int)
        assert expected_balance == balance


@pytest.mark.parametrize(
    "contract, expected_decimals, expected_exception",
    [
        (CONTRACT, 6, None),
        ("TVSvjZdyDSNocHm7dP3jvCmMNsCnMTPa5W", 18, None),
        ("TSos1xxjqMrGKBxycVmtgrnFvv9M6FDFUX", 9, RpcConnectionError),
    ],
)
@vcr_c.use_cassette("tron/get_contract_decimals.yaml")
async def test_get_contract_decimals(tron_client: AioTxTRONClient, contract, expected_decimals, expected_exception):
    if expected_exception:
        with pytest.raises(expected_exception):
            await tron_client.get_contract_decimals(contract)
    else:
        decimals = await tron_client.get_contract_decimals(contract)
        assert decimals == expected_decimals


@vcr_c.use_cassette("tron/get_block_by_number.yaml")
async def test_get_block_by_number(tron_client: AioTxTRONClient):
    block = await tron_client.get_block_by_number(47391985)
    assert isinstance(block, dict)

    assert block["hash"] == "0x0000000002d324f19381c09aa2574e0c1f87a36af924e892e2475338b312ed1b"

    tx_hashes = [tx["hash"] for tx in block["transactions"]]
    assert "0x7262767a999d16cca97807f46e5aac6a8c970b6525e06b49324fb5bfa686bd74" in tx_hashes
    assert "0x2b428b49d2cc2ed2e64e623b570f0478a77aab6d84c5183c200b538a31de017c" in tx_hashes
    assert "0xc69e7d35976989648b88376d82e45a17f35ff441e411cbfc1d3bd86085a10900" in tx_hashes



import decimal
import time
from typing import Optional

from ...boc import Cell, begin_cell, begin_dict
from ...crypto import private_key_to_public_key, verify_sign
from ...utils import Address, sign_message
from .. import Contract
from ._wallet_contract import WalletContract


class MultiSigWalletContractBase(WalletContract):
    def create_data_cell(self):
        owners = begin_dict(8)
        for i in range(len(self.options["public_keys"])):
            owners.store_cell(
                i, begin_cell() \
                    .store_bytes(self.options["public_keys"][i]) \
                    .store_bytes(b'\x00') \
                    .end_cell()
            )

        return begin_cell() \
            .store_uint(self.options["wallet_id"], 32) \
            .store_uint(len(self.options["public_keys"]), 8) \
            .store_uint(self.options["k"], 8) \
            .store_uint(0, 64) \
            .store_maybe_ref(owners.end_cell()) \
            .store_bit(0) \
            .end_cell()


class MultiSigOrder:
    def __init__(self, payload: Cell):
        self.payload = payload
        self.signatures = {}

    def sign(self, owner_id: int, secret_key: bytes):
        signing_hash = self.payload.bytes_hash()
        self.signatures[owner_id] = sign_message(bytes(signing_hash), secret_key).signature
        return signing_hash

    def add_signature(self, owner_id: int, signature: bytes, multisig_wallet):
        signing_hash = self.payload.bytes_hash()
        if not verify_sign(public_key=multisig_wallet.options['public_keys'][owner_id],
                           signed_message=bytes(signing_hash),
                           signature=signature):
            raise Exception('Invalid signature')
        self.signatures[owner_id] = signature

    def union_signatures(self, other):
        self.signatures = dict(list(self.signatures.items()) + list(other.signatures.items()))

    def clear_signatures(self):
        self.signatures = {}

    def to_cell(self, owner_id: int):
        b = begin_cell().store_bit(0)
        for owner in self.signatures:
            signature = self.signatures[owner]
            b = begin_cell() \
                .store_bit(1) \
                .store_ref(
                    begin_cell() \
                    .store_bytes(signature) \
                    .store_uint(owner, 8) \
                    .store_cell(b.end_cell()) \
                    .end_cell()
                )
        return begin_cell() \
               .store_uint(owner_id, 8) \
               .store_cell(b.end_cell()) \
               .store_cell(self.payload) \
               .end_cell()


class MultiSigOrderBuilder:
    def __init__(self, wallet_id, offset=7200, query_id: Optional[int] = None):
        self.wallet_id = wallet_id
        self.messages = begin_cell()
        self.query_id = query_id if query_id is not None else self.generate_query_id(offset)

    def add_message(self, to_addr, amount, payload="", send_mode=3, state_init=None):
        payload_cell = Cell()
        if payload:
            if isinstance(payload, str):
                if len(payload) > 0:
                    payload_cell.bits.write_uint(0, 32)
                    payload_cell.bits.write_string(payload)
            elif hasattr(payload, 'refs'):
                payload_cell = payload
            else:
                payload_cell.bits.write_bytes(payload)

        order_header = Contract.create_internal_message_header(
            Address(to_addr), decimal.Decimal(amount))
        order = Contract.create_common_msg_info(
            order_header, state_init, payload_cell)

        return self.add_message_from_cell(order, send_mode)

    def add_message_from_cell(self, message: Cell, mode: int = 3):
        if len(self.messages.refs) >= 4:
            raise Exception('only 4 refs are allowed')
        self.messages.store_uint(mode, 8)
        self.messages.store_ref(begin_cell().store_cell(message).end_cell())
        return message

    def clear_messages(self):
        self.messages = begin_cell()

    def build(self):
        return MultiSigOrder(begin_cell() \
                             .store_uint(self.wallet_id, 32) \
                             .store_uint(self.query_id, 64) \
                             .store_cell(self.messages.end_cell()) \
                             .end_cell())

    @staticmethod
    def generate_query_id(offset):
        return int(time.time() + offset) << 32


class MultiSigWallet(MultiSigWalletContractBase):
    def __init__(self, **kwargs) -> None:
        # https://github.com/ton-blockchain/multisig-contract/
        # https://github.com/ton-core/ton/blob/master/src/multisig/MultisigWallet.ts
        self.code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
        kwargs["code"] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)
        if "wallet_id" not in kwargs:
            self.options["wallet_id"] = 698983191 + self.options["wc"]

    def get_owner_id_by_public_key(self, public_key: bytes):
        if public_key not in self.options["public_keys"]:
            raise Exception('public key is not an owner')
        return list(self.options["public_keys"]).index(public_key)

    def create_transfer_message(self, order: MultiSigOrder, private_key: bytes, dummy_signature=False):
        public_key = private_key_to_public_key(private_key)
        owner_id = self.get_owner_id_by_public_key(public_key)
        signing_message = order.to_cell(owner_id)

        return self.create_external_message(
            signing_message, private_key, dummy_signature
        )

    def create_external_message(self, signing_message, private_key, dummy_signature=False):
        signature = bytes(64) if dummy_signature else sign_message(
            bytes(signing_message.bytes_hash()), private_key).signature

        body = Cell()
        body.bits.write_bytes(signature)
        body.write_cell(signing_message)

        state_init = code = data = None
        self_address = self.address
        header = Contract.create_external_message_header(self_address)
        result_message = Contract.create_common_msg_info(
            header, state_init, body)

        return {
            "address": self_address,
            "message": result_message,
            "body": body,
            "signature": signature,
            "signing_message": signing_message,
            "state_init": state_init,
            "code": code,
            "data": data,
            "query_id": int.from_bytes(signing_message.bits.array[4:12], 'big')
        }

    def create_init_external_message(self):
        create_state_init = self.create_state_init()
        state_init = create_state_init["state_init"]
        address = create_state_init["address"]
        code = create_state_init["code"]
        data = create_state_init["data"]

        body = Cell()

        header = Contract.create_external_message_header(address)
        external_message = Contract.create_common_msg_info(header, state_init, body)

        return {
            "address": address,
            "message": external_message,
            "body": body,
            "state_init": state_init,
            "code": code,
            "data": data,
        }



from aiotx.clients import AioTxETHClient
import asyncio

eth_client = AioTxETHClient(
    node_url="https://ethereum-sepolia-rpc.publicnode.com")


@eth_client.monitor.on_block
async def handle_block(block):
    print("eth_client: block", block)

@eth_client.monitor.on_transaction
async def handle_transaction(transaction):
    print("eth_client: transaction", transaction)

async def main():
    await eth_client.start_monitoring()
    while True:
        await asyncio.sleep(1)

if __name__ == "__main__":
    asyncio.run(main())


import os
from decimal import Decimal

import pytest
from conftest import vcr_c  # noqa

from aiotx.clients import AioTxBSCClient
from aiotx.exceptions import (
    AioTxError,
    InvalidArgumentError,
    NonceTooLowError,
    ReplacementTransactionUnderpriced,
    TransactionNotFound,
    WrongPrivateKey,
)

PRIVATE_KEY_TO_SEND_FROM = os.environ.get("TEST_BSC_WALLET_PRIVATE_KEY")
assert PRIVATE_KEY_TO_SEND_FROM is not None, "Please provide TEST_BSC_WALLET_PRIVATE_KEY"
CONTRACT = "0x337610d27c682E347C9cD60BD4b3b107C9d34dDd"
DESTINATION_ADDRESS = "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a"


@pytest.mark.parametrize(
    "number, expected_result",
    [
        (5, 5000000000000000000),
        ("0xAB", 171000000000000000000),
        ("0x1E", 30000000000000000000),
        ("0xDEADBEEF", 3735928559000000000000000000),
        (19, 19000000000000000000),
    ],
)
async def test_to_wei(bsc_client: AioTxBSCClient, number, expected_result):
    amount_in_wei = bsc_client.to_wei(number)
    assert expected_result == amount_in_wei


@pytest.mark.parametrize(
    "number, expected_result",
    [
        (5000000000000000000, 5),
        ("0x1234567890ABCDEF12", Decimal('335.812727627494321938')),
        ("0x7236567880ABCDEF12", Decimal('2106.84427382296765621')),
        ("0xDEADBEEF", Decimal('3.735928559E-9')),
        (19000000000000000000, 19),
    ],
)
async def test_from_wei(bsc_client: AioTxBSCClient, number, expected_result):
    amount_in_wei = bsc_client.from_wei(number)
    assert expected_result == amount_in_wei


@vcr_c.use_cassette("bsc/get_last_block.yaml")
async def test_get_last_block(bsc_client: AioTxBSCClient):
    block_id = await bsc_client.get_last_block_number()
    assert isinstance(block_id, int)


@vcr_c.use_cassette("bsc/get_chain_id.yaml")
async def test_get_chain_id(bsc_client: AioTxBSCClient):
    chain_id = await bsc_client.get_chain_id()
    assert isinstance(chain_id, int)
    assert chain_id == 97


@pytest.mark.parametrize(
    "wallet_address, expected_exception, expected_balance",
    [
        ("0x1284214b9b9c85549aB3D2b972df0dEEf66aC2c9", None, 10891723659733231665),
        ("0x3a82a5a2b77d33a12621e860d1d19cdfb8bf63b4", None, 102029280699956200),
        ("0x0e28BD8B2D4efb8A3128f7fB310f4e227E25DB6F", None, 690442684999999988),
        ("0xfb2bdebad0cb9486e714b053a2dbfaf7c05151c9", None, 271930000000000),
        ("0x040bA056435A7675847F4D577d3Cb8Fc2646A946", None, 421930000000000),
        ("0x040bA056435A7675847F4D577d3Cb8Fc2646A948", None, 0),
        ("040bA056435A7675847F4D577d3Cb8Fc2646A946", InvalidArgumentError, 0),
    ],
)
@vcr_c.use_cassette("bsc/get_balance.yaml")
async def test_get_balance(bsc_client: AioTxBSCClient, wallet_address, expected_exception, expected_balance):
    if expected_exception:
        with pytest.raises(expected_exception):
            await bsc_client.get_balance(wallet_address)
    else:
        balance = await bsc_client.get_balance(wallet_address)
        assert isinstance(balance, int)
        assert balance == expected_balance


@pytest.mark.parametrize(
    "tx_id, expected_exception",
    [
        ("0x6007710c9bc82da8a9cb6104e39fef7b259df0d257a0efcf46908b9451909118", None),
        ("0xdeac5f03c42970ec4c89ec540c886a2738b8b8fc8c415d0bd231ea94dad81727", None),
        ("0x26611f5796882c8f45a52c4443c2cda4580ef0bb06d57686d34672207312dfd3", None),
        ("0xf82c96662db872899f16ea9dfed73fb7c29f3942fdbfc97500ca071793f5a397", None),
        ("0x8a0dc46462919c8a1c15c90e79782ba4e56b9053fad8975c9b6bec094f001895", None),
        ("0x41c8e2e03195c29cea37d67ad234d91c52258a514c78f9fcfd196aa5992209ae", None),
        ("0x6007710c9bc82da8a9cb6104e39fef7b259df0d257a0efcf46908b9451909111", TransactionNotFound),
        ("0x6007710c9bc82da8a9cb6104e39fef7b259df0d257a0efcf46908b945190911", InvalidArgumentError),
        ("41c8e2e03195c29cea37d67ad234d91c52258a514c78f9fcfd196aa5992209ae", InvalidArgumentError),
    ],
)
@vcr_c.use_cassette("bsc/get_transaction.yaml")
async def test_get_transaction(bsc_client: AioTxBSCClient, tx_id, expected_exception):
    if expected_exception:
        with pytest.raises(expected_exception):
            await bsc_client.get_transaction(tx_id)
    else:
        tx = await bsc_client.get_transaction(tx_id)
        assert "aiotx_decoded_input" in tx.keys()


@pytest.mark.parametrize(
    "wallet_address, contract, expected_exception, expected_balance",
    [
        ("0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a", CONTRACT, None, 8000000001000000000),
        ("0x896c54023265f2c821f241BC694d8A0E1FA7DF2E", CONTRACT, None, 3570000000000000000),
        ("0x3ffeCb152F95f7122990ab16Eff4B0B5d533497e", CONTRACT, None, 8469999999000000000),
        ("0x98265f1Dd3224aBA2148C6cF2Ef873b05A3476C1", CONTRACT, None, 659546202581319287241),
        ("0xf6626A900e4Cc958D2cD0Eb2186Fd6B29EDB63ce", CONTRACT, None, 950000000000000000),
        ("0x2dbB5a4c235164B9f772179A43faca2c71a8abDB", CONTRACT, None, 41624579772130211939),
        ("0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C", CONTRACT, None, 0),
        ("0xDA31b5ac94C559478Eeaa7E347ee4557f9a6F71C", "Eeaa7E347ee4557f9a6F71C", InvalidArgumentError, 0),
    ],
)
@vcr_c.use_cassette("bsc/get_token_balance.yaml")
async def test_get_token_balance(
    bsc_client: AioTxBSCClient, wallet_address, contract, expected_exception, expected_balance
):

    if expected_exception:
        with pytest.raises(expected_exception):
            await bsc_client.get_contract_balance(wallet_address, contract)
    else:
        balance = await bsc_client.get_contract_balance(wallet_address, contract)
        assert isinstance(balance, int)
        assert expected_balance == balance


@pytest.mark.parametrize(
    "wallet_address, expected_exception, expected_count",
    [
        ("0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a", None, 0),
        ("0x896c54023265f2c821f241BC694d8A0E1FA7DF2E", None, 0),
        ("0x3ffeCb152F95f7122990ab16Eff4B0B5d533497e", None, 132),
        ("0x98265f1Dd3224aBA2148C6cF2Ef873b05A3476C1", None, 1788),
        ("0xf6626A900e4Cc958D2cD0Eb2186Fd6B29EDB63ce", None, 17),
        ("0x2dbB5a4c235164B9f772179A43faca2c71a8abDB", None, 1),
        ("2dbB5a4c235164B9f772179A43faca2c71a8abDB", InvalidArgumentError, 0),
        ("0x2dbB5a4c235164B9f772179A43faca2c71a8ab", InvalidArgumentError, 0),
    ],
)
@vcr_c.use_cassette("bsc/get_transaction_count.yaml")
async def test_get_transaction_count(bsc_client: AioTxBSCClient, wallet_address, expected_exception, expected_count):
    if expected_exception:
        with pytest.raises(expected_exception):
            await bsc_client.get_transactions_count(wallet_address)
    else:
        count = await bsc_client.get_transactions_count(wallet_address)
        assert isinstance(count, int)
        assert expected_count == count


@vcr_c.use_cassette("bsc/get_gas_price.yaml")
async def test_send_get_gas_price(bsc_client: AioTxBSCClient):
    result = await bsc_client.get_gas_price()
    assert isinstance(result, int)
    assert result == 5050000000


@pytest.mark.parametrize(
    "private_key, to_address, amount, gas_price, gas_limit, expected_exception",
    [
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            WrongPrivateKey,
        ),
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938ch",
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            WrongPrivateKey,
        ),
        (
            "e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            0.00001,
            5,
            21000,
            WrongPrivateKey,
        ),
        (PRIVATE_KEY_TO_SEND_FROM, "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5", 0.00001, 5, 21000, ValueError),
        (PRIVATE_KEY_TO_SEND_FROM, "f9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a", 0.00001, 5, 21000, None),
        (PRIVATE_KEY_TO_SEND_FROM, "0xf9E35E4e1CbcF08E84d3f6FF662Ba4c306b5a", 0.00001, 5, 21000, ValueError),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 0.00001, 5, 21000, None),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 5, 5, 21000, AioTxError),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 0.00001, 0, 21000, AioTxError),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 0, 5, 21000, ReplacementTransactionUnderpriced),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 0.00001, 5, 0, AioTxError),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 0.00001, 5, 61000, ReplacementTransactionUnderpriced),
    ],
)
@vcr_c.use_cassette("bsc/send_transaction.yaml")
async def test_send_transaction(
    bsc_client: AioTxBSCClient, private_key, to_address, amount, gas_price, gas_limit, expected_exception
):
    """
    Here it's raising ReplacementTransactionUnderpriced and NonceTooLowError because we have reusing
    the same VCR data for every get nonce request, we should investigate how we can change that maybe?
    """
    gas_price = bsc_client.to_wei(gas_price, "gwei")
    wei_amount = bsc_client.to_wei(amount, "ether")
    if expected_exception:
        with pytest.raises(expected_exception):
            await bsc_client.send(private_key, to_address, wei_amount, gas_price=gas_price, gas_limit=gas_limit)
    else:
        result = await bsc_client.send(private_key, to_address, wei_amount, gas_price=gas_price, gas_limit=gas_limit)
        assert isinstance(result, str)


@pytest.mark.parametrize(
    "private_key, to_address, amount, expected_exception",
    [
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, 0.00001, None),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            "0x3ffeCb152F95f7122990ab16Eff4B0B5d533497e",
            0.00001,
            ReplacementTransactionUnderpriced,
        ),
    ],
)
@vcr_c.use_cassette("bsc/send_transaction_with_auto_gas.yaml")
async def test_send_transaction_with_auto_gas(
    bsc_client: AioTxBSCClient, private_key, to_address, amount, expected_exception
):
    """
    Here it's raising ReplacementTransactionUnderpriced and NonceTooLowError because we have reusing
    the same VCR data for every get nonce request, we should investigate how we can change that maybe?
    """
    wei_amount = bsc_client.to_wei(amount, "ether")
    if expected_exception:
        with pytest.raises(expected_exception):
            await bsc_client.send(private_key, to_address, wei_amount)
    else:
        result = await bsc_client.send(private_key, to_address, wei_amount)
        assert isinstance(result, str)


@vcr_c.use_cassette("bsc/send_transaction_with_custom_nonce.yaml")
async def test_send_transaction_with_custom_nonce(bsc_client: AioTxBSCClient):
    wei_amount = bsc_client.to_wei(0.00002, "ether")
    sender_address = bsc_client.get_address_from_private_key(PRIVATE_KEY_TO_SEND_FROM)
    nonce = await bsc_client.get_transactions_count(sender_address)
    first_tx = await bsc_client.send(PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, wei_amount, nonce=nonce)
    assert isinstance(first_tx, str)
    second_tx = await bsc_client.send(PRIVATE_KEY_TO_SEND_FROM, sender_address, wei_amount, nonce=nonce + 1)
    assert isinstance(second_tx, str)


@pytest.mark.parametrize(
    "private_key, to_address, contract, amount, gas_price, gas_limit, expected_exception",
    [
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            WrongPrivateKey,
        ),
        (
            "87e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938ch",
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            WrongPrivateKey,
        ),
        (
            "e6dah15aa076a932cd9f0663da72f8cfb6d3e23c00ef1269104bd904938chd",
            DESTINATION_ADDRESS,
            CONTRACT,
            1,
            5,
            61000,
            WrongPrivateKey,
        ),
        (PRIVATE_KEY_TO_SEND_FROM, "0xf9E35E4e1CbcF08E984d3f6FF662Ba4c306b5a", CONTRACT, 1, 5, 61000, ValueError),
        (PRIVATE_KEY_TO_SEND_FROM, "f9E5E4e1CbcF08E99B84d3f6FF662Ba4c306b5a", CONTRACT, 1, 5, 61000, ValueError),
        (PRIVATE_KEY_TO_SEND_FROM, "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b", CONTRACT, 1, 5, 61000, ValueError),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            "0xf9E35E4e1CbcF08E984d3f6FF662Ba4c306b5a",
            0.00001,
            5,
            61000,
            ValueError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            "f9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b5a",
            1,
            5,
            61000,
            NonceTooLowError,
        ),
        (
            PRIVATE_KEY_TO_SEND_FROM,
            DESTINATION_ADDRESS,
            "0xf9E35E4e1CbcF08E99B84d3f6FF662Ba4c306b",
            1,
            5,
            61000,
            ValueError,
        ),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, CONTRACT, 0.5, 5, 61000, None),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, CONTRACT, 1000, 5, 61000, ReplacementTransactionUnderpriced),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, CONTRACT, 1, 0, 61000, AioTxError),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, CONTRACT, 0, 5, 61000, ReplacementTransactionUnderpriced),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, CONTRACT, 1, 5, 0, AioTxError),
        (PRIVATE_KEY_TO_SEND_FROM, DESTINATION_ADDRESS, CONTRACT, 0.5, 5, 61000, None),
    ],
)
@vcr_c.use_cassette("bsc/send_token_transaction.yaml")
async def test_send_token_transaction(
    bsc_client: AioTxBSCClient, private_key, to_address, contract, amount, gas_price, gas_limit, expected_exception
):
    """
    Here it's raising ReplacementTransactionUnderpriced and NonceTooLowError because we have reusing
    the same VCR data for every get nonce request, we should investigate how we can change that maybe?
    """
    gas_price = bsc_client.to_wei(gas_price, "gwei")
    wei_amount = bsc_client.to_wei(amount, "ether")

    if expected_exception:
        with pytest.raises(expected_exception):
            await bsc_client.send_token(
                private_key, to_address, contract, wei_amount, gas_price=gas_price, gas_limit=gas_limit
            )
    else:
        result = await bsc_client.send_token(
            private_key, to_address, contract, wei_amount, gas_price=gas_price, gas_limit=gas_limit
        )
        assert isinstance(result, str)


@pytest.mark.parametrize(
    "contract, expected_decimals, expected_exception",
    [
        (CONTRACT, 18, None),
        ("0x337610d27c682E347C9cD60BD4b3b107C9d34dD", 18, InvalidArgumentError),
        ("0x757fc78bb4df4ce6605ae669bf0b71074176aac3", 9, None),
    ],
)
@vcr_c.use_cassette("bsc/get_contract_decimals.yaml")
async def test_get_contract_decimals(bsc_client: AioTxBSCClient, contract, expected_decimals, expected_exception):
    if expected_exception:
        with pytest.raises(expected_exception):
            await bsc_client.get_contract_decimals(contract)
    else:
        decimals = await bsc_client.get_contract_decimals(contract)
        assert decimals == expected_decimals



import decimal
import time

from ...boc import Cell, begin_cell, begin_dict
from ...utils import Address, sign_message
from .. import Contract
from ._wallet_contract import WalletContract


class HighloadWalletContractBase(WalletContract):
    def create_data_cell(self):
        return begin_cell() \
            .store_uint(self.options["wallet_id"], 32) \
            .store_uint(0, 64) \
            .store_bytes(self.options["public_key"]) \
            .store_maybe_ref(None) \
            .end_cell()

    def create_signing_message(self, query_id: int=0):
        message = begin_cell().store_uint(self.options["wallet_id"], 32)
        return message.store_uint(query_id, 64)


class HighloadWalletV2Contract(HighloadWalletContractBase):
    def __init__(self, **kwargs) -> None:
        # https://github.com/akifoq/highload-wallet/blob/master/highload-wallet-v2-code.fc
        self.code = "B5EE9C720101090100E5000114FF00F4A413F4BCF2C80B010201200203020148040501EAF28308D71820D31FD33FF823AA1F5320B9F263ED44D0D31FD33FD3FFF404D153608040F40E6FA131F2605173BAF2A207F901541087F910F2A302F404D1F8007F8E16218010F4786FA5209802D307D43001FB009132E201B3E65B8325A1C840348040F4438AE63101C8CB1F13CB3FCBFFF400C9ED54080004D03002012006070017BD9CE76A26869AF98EB85FFC0041BE5F976A268698F98E99FE9FF98FA0268A91040207A0737D098C92DBFC95DD1F140034208040F4966FA56C122094305303B9DE2093333601926C21E2B3"
        kwargs["code"] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)
        if "wallet_id" not in kwargs:
            self.options["wallet_id"] = 698983191 + self.options["wc"]

    def create_transfer_message(self, recipients_list: list, query_id: int, timeout=60, dummy_signature=False):
        if query_id < int(time.time() + timeout) << 32:
            query_id = int(time.time() + timeout) << 32 + query_id

        signing_message = self.create_signing_message(query_id)
        recipients = begin_dict(16)
        for i, recipient in enumerate(recipients_list):
            payload_cell = Cell()
            if recipient.get('payload'):
                if isinstance(recipient['payload'], str):
                    if len(recipient['payload']) > 0:
                        payload_cell.bits.write_uint(0, 32)
                        payload_cell.bits.write_string(recipient['payload'])
                elif hasattr(recipient['payload'], 'refs'):
                    payload_cell = recipient['payload']
                else:
                    payload_cell.bits.write_bytes(recipient['payload'])

            order_header = Contract.create_internal_message_header(
                Address(recipient['address']), decimal.Decimal(recipient['amount'])
            )
            order = Contract.create_common_msg_info(
                order_header, recipient.get('state_init'), payload_cell
            )
            recipients.store_cell(
                i, begin_cell() \
                    .store_uint8(recipient.get('send_mode', 0)) \
                    .store_ref(order).end_cell()
            )

        signing_message.store_maybe_ref(recipients.end_cell())
        return self.create_external_message(
            signing_message.end_cell(), dummy_signature
        )

    def create_external_message(self, signing_message, dummy_signature=False):
        signature = bytes(64) if dummy_signature else sign_message(
            bytes(signing_message.bytes_hash()), self.options['private_key']).signature

        body = Cell()
        body.bits.write_bytes(signature)
        body.write_cell(signing_message)

        state_init = code = data = None
        self_address = self.address
        header = Contract.create_external_message_header(self_address)
        result_message = Contract.create_common_msg_info(
            header, state_init, body)

        return {
            "address": self_address,
            "message": result_message,
            "body": body,
            "signature": signature,
            "signing_message": signing_message,
            "state_init": state_init,
            "code": code,
            "data": data,
            "query_id": int.from_bytes(signing_message.bits.array[4:12], 'big')
        }

    def create_init_external_message(self, timeout=60):
        create_state_init = self.create_state_init()
        state_init = create_state_init["state_init"]
        address = create_state_init["address"]
        code = create_state_init["code"]
        data = create_state_init["data"]

        signing_message = self.create_signing_message(int(time.time() + timeout) << 32) \
            .store_maybe_ref(None).end_cell()
        signature = sign_message(
            bytes(signing_message.bytes_hash()), self.options['private_key']).signature

        body = Cell()
        body.bits.write_bytes(signature)
        body.write_cell(signing_message)

        header = Contract.create_external_message_header(address)
        external_message = Contract.create_common_msg_info(
            header, state_init, body)

        return {
            "address": address,
            "message": external_message,

            "body": body,
            "signing_message": signing_message,
            "state_init": state_init,
            "code": code,
            "data": data,
        }



from ._builder import Builder, begin_cell
from ._cell import Cell, deserialize_cell_data, parse_boc_header
from ._dict_builder import DictBuilder, begin_dict

__all__ = [
    'Cell',
    'Builder', 'begin_cell',
    'DictBuilder', 'begin_dict',
    'deserialize_cell_data',
    'parse_boc_header',
]



from math import ceil, log2

from .._cell import Cell
from .find_common_prefix import find_common_prefix


def pad(src: str, size: int) -> str:
    while len(src) < size:
        src = '0' + src

    return src


def remove_prefix_map(src, length):
    if length == 0:
        return src
    else:
        res = {}
        for k in src:
            res[k[length:]] = src[k]

        return res


def fork_map(src):
    assert len(src) > 0, 'Internal inconsistency'
    left = {}
    right = {}
    for k in src:
        if k.find('0') == 0:
            left[k[1:]] = src[k]
        else:
            right[k[1:]] = src[k]

    assert len(left) > 0, 'Internal inconsistency. Left empty.'
    assert len(right) > 0, 'Internal inconsistency. Left empty.'
    return left, right


def build_node(src):
    assert len(src) > 0, 'Internal inconsistency'
    if len(src) == 1:
        return {
            'type': 'leaf',
            'value': list(src.values())[0]
        }

    left, right = fork_map(src)
    return {
        'type': 'fork',
        'left': build_edge(left),
        'right': build_edge(right)
    }


def build_edge(src):
    assert len(src) > 0, 'Internal inconsistency'
    label = find_common_prefix(list(src.keys()))
    return {
        'label': label,
        'node': build_node(
            remove_prefix_map(src, len(label))
        )
    }


def build_tree(src, key_size):
    # Convert map keys
    tree = {}
    for key in src:
        padded = pad(bin(key)[2:], key_size)
        tree[padded] = src[key]

    # Calculate root label
    return build_edge(tree)


# Serialization
def write_label_short(src, to):
    # Header
    to.write_bit(0)

    # Unary length
    for e in src: 
        to.write_bit(1)
    to.write_bit(0)

    # Value
    for e in src:
        to.write_bit(e == '1')

    return to


def label_short_length(src):
    return 1 + len(src) + 1 + len(src)


def write_label_long(src, key_length, to):
    # Header
    to.write_bit(1)
    to.write_bit(0)

    # Length
    length = ceil(log2(key_length + 1))
    to.write_uint(len(src), length)

    # Value
    for e in src:
        to.write_bit(e == '1')

    return to


def label_long_length(src, key_length):
    return 1 + 1 + ceil(log2(key_length + 1)) + len(src)


def write_label_same(value: bool, length, key_length, to):
    to.write_bit(1)
    to.write_bit(1)

    to.write_bit(value)

    len_len = ceil(log2(key_length + 1))
    to.write_uint(length, len_len)


def label_same_length(key_size):
    return 1 + 1 + 1 + ceil(log2(key_size + 1))


def is_same(src):
    if len(src) == 0 or len(src) == 1:
        return True

    for e in src[1:]:
        if e != src[0]:
            return False

    return True


def detect_label_type(src, key_size):
    kind = 'short'
    kind_length = label_short_length(src)

    long_length = label_long_length(src, key_size)
    if long_length < kind_length:
        kind_length = long_length
        kind = 'long'

    if is_same(src):
        same_length = label_same_length(key_size)
        if same_length < kind_length:
            kind_length = same_length
            kind = 'same'

    return kind


def write_label(src, key_size, to):
    type = detect_label_type(src, key_size)
    if type == 'short':
        write_label_short(src, to)
    elif type == 'long':
        write_label_long(src, key_size, to)
    elif type == 'same':
        write_label_same(src[0] == '1', len(src), key_size, to)


def write_node(src, key_size, serializer, to):
    if src['type'] == 'leaf':
        serializer(src['value'], to)

    if src['type'] == 'fork':
        left_cell = Cell()
        right_cell = Cell()
        write_edge(src['left'], key_size - 1, serializer, left_cell)
        write_edge(src['right'], key_size - 1, serializer, right_cell)
        to.refs.append(left_cell)
        to.refs.append(right_cell)


def write_edge(src, key_size, serializer, to):
    write_label(src['label'], key_size, to.bits)
    write_node(src['node'], key_size - len(src['label']), serializer, to)


def serialize_dict(src, key_size, serializer):
    tree = build_tree(src, key_size)
    dest = Cell()
    write_edge(tree, key_size, serializer, dest)
    return dest



from aiotx.clients import AioTxLTCClient
import asyncio

ltc_client = AioTxLTCClient(node_url="https://api.tatum.io/v3/blockchain/node/litecoin-core-testnet", testnet=True)


@ltc_client.monitor.on_block
async def handle_block(block):
    print("ltc_client: block", block)

@ltc_client.monitor.on_transaction
async def handle_transaction(transaction):
    print("ltc_client: transaction", transaction)

async def main():
    await ltc_client.start_monitoring()
    while True:
        await asyncio.sleep(1)

if __name__ == "__main__":
    asyncio.run(main())


from enum import Enum


class BlockParam(Enum):
    LATEST = "latest"
    EARLIEST = "earliest"
    PENDING = "pending"
    SAFE = "safe"
    FINALIZED = "finalized"


class FeeEstimate(Enum):
    UNSET = "UNSET"
    ECONOMICAL = "ECONOMICAL"
    CONSERVATIVE = "CONSERVATIVE"



from setuptools import find_packages, setup


extras_test = (
    [
        "hypothesis",
        "ruff",
        "pyproj",
        "pytest",
        "pytest-cov",
        "sphinx",
        "pytest-asyncio",
        "tox",
        "build",
        "vcrpy"
    ]
)

setup(
    name="aiotx",
    keywords=[
        "cryptocurrency",
        "blockchain",
        "bitcoin",
        "ethereum",
        "asyncio",
        "aiohttp",
        "json-rpc",
        "payment",
        "wallet",
        "transaction",
    ],
    use_scm_version=True,
    description="An asynchronous library for interacting with various cryptocurrencies and blockchains",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests", "scripts", "examples", "docs")),
    package_data={
        "aiotx": ["py.typed"],
        "aiotx.utils": ["*.json"],
    },
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=[
        "aiohttp",
        "eth_keys==0.5.*",
        "eth_utils==4.1.*",
        "eth_account==0.13.*",
        "sqlalchemy",
        "aiosqlite",
        "aiomysql",
        "cryptography",
        "bitcoinlib==0.6.*",
        "greenlet",
        "setuptools",
        "pynacl>=1.4.0",
        "tronpy",
    ],
    extras_require={
        "test": extras_test,
    },
    url="https://github.com/Grommash9/aiotx",
    project_urls={
        'Documentation': 'https://grommash9.github.io/aiotx/',
        'Source': 'https://github.com/Grommash9/aiotx'
        },
    author="Oleksandr Prudnikov",
    author_email="prudnikov21@icloud.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)



import copy
import math
from typing import Optional, Union

from ..utils._address import Address


class BitString:
    def __init__(self, length: int):
        self.array = bytearray(math.ceil(length / 8))
        self.cursor = 0
        self.length = length

    def __repr__(self):
        return str(self.get_top_upped_array())

    def __iter__(self):
        for i in range(self.cursor):
            yield self.get(i)

    def __getitem__(self, key):
        if isinstance(key, slice):
            start = key.start if key.start else 0
            stop = key.stop if key.stop else len(self)
            step = key.step if key.step else 1

            return [self[ii] for ii in range(start, stop, step)]
        elif isinstance(key, int):
            if key < 0:
                key += len(self)
            if key < 0 or key >= len(self):
                raise IndexError("The index (%d) is out of range." % key)
            return self.get(key)
        else:
            raise TypeError("Invalid argument type.")

    def __len__(self):
        return self.length

    def get(self, n: int):
        """Just returns n bits from cursor. Does not move the cursor."""
        return int((self.array[(n // 8) | 0] & (1 << (7 - (n % 8)))) > 0)

    def off(self, n):
        """Sets next from cursor n bits to 0. Does not move cursor."""
        self.check_range(n)
        self.array[(n // 8) | 0] &= ~(1 << (7 - (n % 8)))

    def on(self, n):
        """Sets next from cursor n bits to 1. Does not move cursor."""
        self.check_range(n)
        self.array[(n // 8) | 0] |= 1 << (7 - (n % 8))

    def check_range(self, n: int) -> None:
        """Throws an exception if the cursor + n is out of range."""
        if n > self.length:
            raise Exception("BitString overflow")

    def set_top_upped_array(self, array: bytearray, fullfilled_bytes=True):
        self.length = len(array) * 8
        self.array = array
        self.cursor = self.length

        if fullfilled_bytes or not self.length:
            return

        else:
            found_end_bit = False
            for c in range(7):
                self.cursor -= 1

                if self.get(self.cursor):
                    found_end_bit = True
                    self.off(self.cursor)
                    break

            if not found_end_bit:
                raise Exception(
                    f"Incorrect TopUppedArray {array}, {fullfilled_bytes}")

    def get_top_upped_array(self) -> bytearray:
        ret = copy.deepcopy(self)
        tu = math.ceil(ret.cursor / 8) * 8 - ret.cursor
        if tu > 0:
            tu -= 1
            ret.write_bit(1)
            while tu > 0:
                tu -= 1
                ret.write_bit(0)
        ret.array = ret.array[:math.ceil(ret.cursor / 8)]
        return ret.array

    def get_free_bits(self) -> int:
        """Returns the number of not used bits in the BitString."""
        return self.length - self.cursor

    def get_used_bits(self):
        return self.cursor

    def write_bit_array(self, ba: bytearray):
        """Writes a bytearray as a bit array one bit by one."""
        for b in ba.decode('utf-8'):
            self.write_bit(b)

    def write_bit(self, b: Union[str, int]):
        b = int(b)
        if b == 1:
            self.on(self.cursor)
        elif b == 0:
            self.off(self.cursor)
        else:
            raise Exception("BitString can only write 1 or 0")

        self.cursor += 1

    def write_uint(self, number: int, bit_length: int):
        if bit_length == 0 or len("{0:b}".format(number)) > bit_length:
            if number == 0:
                return

            raise Exception(
                f"bitLength is too small for number, got number={number},bitLength={bit_length}")

        for i in range(bit_length, 0, -1):
            k = (2 ** (i - 1))
            if number // k == 1:
                self.write_bit(1)
                number -= k
            else:
                self.write_bit(0)

    def write_uint8(self, ui8: int):
        """Just as write_uint(n, 8), but only write_uint8(n) (?)."""
        self.write_uint(ui8, 8)

    def write_int(self, number: int, bit_length: int):
        if bit_length == 1:
            if number == -1:
                self.write_bit(1)
                return

            if number == 0:
                self.write_bit(0)
                return

            raise Exception("Bitlength is too small for number")
        else:
            if number < 0:
                self.write_bit(1)
                s = 2 ** (bit_length - 1)
                self.write_uint(s + number, bit_length - 1)
            else:
                self.write_bit(0)
                self.write_uint(number, bit_length - 1)

    def write_string(self, value: str):
        self.write_bytes(bytes(value, encoding="utf-8"))

    def write_bytes(self, ui8_array: bytes):
        for ui8 in ui8_array:
            self.write_uint8(ui8)

    def write_bit_string(self, another_bit_string: "BitString"):
        for bit in another_bit_string:
            self.write_bit(bit)

    def write_address(self, address: Optional[Address]):
        """Writes an address, maybe zero-address (None) to the BitString."""
        if address is None:
            self.write_uint(0, 2)
        else:
            self.write_uint(2, 2)
            self.write_uint(0, 1)  # anycast
            self.write_int(address.wc, 8)
            self.write_bytes(address.hash_part)

    def write_grams(self, amount: int):
        if amount == 0:
            self.write_uint(0, 4)
        else:
            amount = int(amount)
            length = math.ceil(len(hex(amount)[2:]) / 2)  # ? [2:] removes 0x
            self.write_uint(length, 4)
            self.write_uint(amount, length * 8)

    def write_coins(self, amount):
        self.write_grams(amount)



# https://github.com/vergl4s/ethereum-mnemonic-utils/blob/master/mnemonic_utils.py
import math
import os
from hashlib import pbkdf2_hmac

from nacl.bindings import crypto_sign_ed25519_sk_to_pk
from nacl.signing import VerifyKey, exc

from ._settings import PBKDF_ITERATIONS


def get_secure_random_number(min_v, max_v):
    range_betw = max_v - min_v
    bits_needed = math.ceil(math.log2(range_betw))
    if bits_needed > 53:
        raise Exception("Range is too large")

    bytes_needed = math.ceil(bits_needed / 8)
    mask = math.pow(2, bits_needed) - 1

    while True:
        res = os.urandom(bits_needed)
        power = (bytes_needed - 1) * 8
        number_val = 0
        for i in range(bytes_needed):
            number_val += res[i] * math.pow(2, power)
            power -= 8
        number_val = int(number_val) & int(mask)
        if number_val >= range_betw:
            continue

        return min_v + number_val


def is_basic_seed(entropy):
    seed = pbkdf2_hmac("sha512", entropy, 'TON seed version'.encode(
        'utf-8'), max(1, math.floor(PBKDF_ITERATIONS / 256)))
    return seed[0] == 0


def private_key_to_public_key(priv_k: bytes):
    return crypto_sign_ed25519_sk_to_pk(priv_k)


def verify_sign(public_key: bytes, signed_message: bytes, signature: bytes):
    key = VerifyKey(public_key)
    try:
        key.verify(signed_message, signature)
        return True
    except exc.BadSignatureError:
        return False



import time

from ...boc import Cell
from ._wallet_contract import WalletContract


class WalletV3ContractBase(WalletContract):
    def create_data_cell(self):
        cell = Cell()
        cell.bits.write_uint(0, 32)
        cell.bits.write_uint(self.options["wallet_id"], 32)
        cell.bits.write_bytes(self.options["public_key"])
        return cell

    def create_signing_message(self, seqno=None):
        seqno = seqno or 0
        message = Cell()
        message.bits.write_uint(self.options["wallet_id"], 32)
        if seqno == 0:
            for _ in range(32):
                message.bits.write_bit(1)
        else:
            timestamp = int(time.time())  # get timestamp in seconds
            message.bits.write_uint(timestamp + 60, 32)

        message.bits.write_uint(seqno, 32)
        return message


class WalletV3ContractR1(WalletV3ContractBase):
    def __init__(self, **kwargs) -> None:
        self.code = "B5EE9C724101010100620000C0FF0020DD2082014C97BA9730ED44D0D70B1FE0A4F2608308D71820D31FD31FD31FF82313BBF263ED44D0D31FD31FD3FFD15132BAF2A15144BAF2A204F901541055F910F2A3F8009320D74A96D307D402FB00E8D101A4C8CB1FCB1FCBFFC9ED543FBE6EE0"
        kwargs["code"] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)
        if "wallet_id" not in kwargs:
            self.options["wallet_id"] = 698983191 + self.options["wc"]


class WalletV3ContractR2(WalletV3ContractBase):
    def __init__(self, **kwargs) -> None:
        self.code = "B5EE9C724101010100710000DEFF0020DD2082014C97BA218201339CBAB19F71B0ED44D0D31FD31F31D70BFFE304E0A4F2608308D71820D31FD31FD31FF82313BBF263ED44D0D31FD31FD3FFD15132BAF2A15144BAF2A204F901541055F910F2A3F8009320D74A96D307D402FB00E8D101A4C8CB1FCB1FCBFFC9ED5410BD6DAD"
        kwargs["code"] = Cell.one_from_boc(self.code)
        super().__init__(**kwargs)
        if "wallet_id" not in kwargs:
            self.options["wallet_id"] = 698983191 + self.options["wc"]



import logging

logger = logging.getLogger(__name__)

class ColoredFormatter(logging.Formatter):
    COLOR_CODES = {
        logging.CRITICAL: "\033[1;35m",  # Magenta
        logging.ERROR: "\033[1;31m",     # Red
        logging.WARNING: "\033[1;33m",   # Yellow
        logging.INFO: "\033[0;37m",      # White
        logging.DEBUG: "\033[1;34m",     # Blue
    }

    def format(self, record):
        color = self.COLOR_CODES.get(record.levelno, "")
        reset = "\033[0m"
        formatted = super().format(record)
        return f"{color}{formatted}{reset}"
    

def set_logger_level(level):
    """
    Set the logger level.

    Args:
        level (str): The desired logger level.
            Possible values: "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
    """
    try:
        logger.setLevel(getattr(logging, level.upper()))
    except AttributeError:
        valid_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        raise ValueError(f"Invalid logger level: {level}. Valid levels are: {', '.join(valid_levels)}")

handler = logging.StreamHandler()
handler.setFormatter(ColoredFormatter())
logger.addHandler(handler)


from aiotx.clients import AioTxTRONClient
import asyncio

tron_client = AioTxTRONClient(node_url="https://go.getblock.io/<token>/jsonrpc")


@tron_client.monitor.on_block
async def handle_block(block):
    print("tron_client: block", block)

@tron_client.monitor.on_transaction
async def handle_transaction(transaction):
    print("tron_client: transaction", transaction)

async def main():
    await tron_client.start_monitoring()
    while True:
        await asyncio.sleep(1)

if __name__ == "__main__":
    asyncio.run(main())